/**
 * @file i2c_usb.c
 *
 * @brief This file contains the command processing for the I2C-MP-USB firmware
 *        project. This code is based on the I2C-Tiny-USB project by Till
 *        Harbaum (https://github.com/harbaum/I2C-Tiny-USB).
 *
 * @author Thomas Fischl
 * @copyright (c) 2017-2019 Thomas Fischl
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "i2c.h"
#include "led.h"
#include "bootloader.h"
#include "i2c_usb.h"

unsigned char status = STATUS_IDLE;
unsigned short expected = 0;
unsigned char saved_cmd;

/**
 * @brief Process given i2c command
 * 
 * @param cmd I2C command to process
 * @return 0xff if there are more data to send, else 0x00
 */
unsigned char i2c_do(struct i2c_cmd *cmd) {

    unsigned char address = (cmd->addr << 1);
    if (cmd->flags & I2C_M_RD)
        address |= 1;

    if (cmd->cmd & CMD_I2C_BEGIN)
        i2c_start();
    else
        i2c_repeatedStart();


    // send DEVICE address
    if (!i2c_sendByte(address)) {

        status = STATUS_ADDRESS_NAK;
        expected = 0;
        i2c_stop();

    } else {

        status = STATUS_ADDRESS_ACK;
        expected = cmd->len;
        saved_cmd = cmd->cmd;

        /* check if transfer is already done (or failed) */
        if ((cmd->cmd & CMD_I2C_END) && !expected)
            i2c_stop();
    }


    return (cmd->len ? 0xff : 0x00);
}

/**
 * @brief Handle the USB setup packets
 * 
 * @param data USB packet data for request
 * @param response USB packet data for response
 * @return Length of data to send out
 */
unsigned char usb_vendor_setup(unsigned char * data, unsigned char * response) {

    switch (data[1]) {

        case CMD_ECHO: // echo (for transfer reliability testing)
            response[0] = data[2];
            response[1] = data[3];
            return 2;

        case CMD_GET_FUNC:
            response[0] = FUNCTION_CAPABILITY & 0xff;
            response[1] = (FUNCTION_CAPABILITY >> 8) & 0xff;
            response[2] = (FUNCTION_CAPABILITY >> 16) & 0xff;
            response[3] = (FUNCTION_CAPABILITY >> 24) & 0xff;
            return 4;

        case CMD_SET_DELAY:
            i2c_setBaudratePeriod((unsigned short) data[2] + ((unsigned short) data[3] << 8));
            return 0;

        case CMD_GET_STATUS:
            response[0] = status;
            return 1;

        case CMD_I2C_IO:
        case CMD_I2C_IO + CMD_I2C_BEGIN:
        case CMD_I2C_IO + CMD_I2C_END:
        case CMD_I2C_IO + CMD_I2C_BEGIN + CMD_I2C_END:
            led_flash();
            return i2c_do((struct i2c_cmd*) data);

        case CMD_START_BOOTLOADER:
            if ((data[2] == 0x37) && (data[3] == 0x52))
                bootloader_triggerStart();
            return 0;
            
        case CMD_SET_BAUDRATE:
            i2c_setBaudrate((unsigned short) data[2] + ((unsigned short) data[3] << 8));
            return 0;
            
            
    }

    return 0;
}

/**
 * @brief Handle USB in requests
 * 
 * @param data Pointer to packet data to fill
 * @param length Count of maximum data length
 * @return Count of written data
 */
unsigned char usb_vendor_in(unsigned char * data, unsigned char length) {

    unsigned char i;

    if (status == STATUS_ADDRESS_ACK) {


        if (length > expected) {
            length = expected;
        }

        // consume bytes
        for (i = 0; i < length; i++) {
            expected--;
            i2c_receiveByte(&data[i], expected == 0); // TODO: handle return value
        }

        // end transfer on last byte
        if ((saved_cmd & CMD_I2C_END) && !expected)
            i2c_stop();

    } else {

        for (i = 0; i < length; i++) {
            data[i] = 0;
        }
    }

    return length;
}

/**
 * @brief Handle USB out requests
 * 
 * @param data Byte field containing USB data packet
 * @param length Length of data packet
 * @return 1 if ack should be sent
 */
unsigned char usb_vendor_out(unsigned char * data, unsigned char length) {

    unsigned char i;
    unsigned char err = 0;

    if (status == STATUS_ADDRESS_ACK) {

        if (length > expected) {
            length = expected;
        }

        // consume bytes
        for (i = 0; i < length; i++) {
            expected--;
            if (!i2c_sendByte(data[i]))
                err = 1;
        }

        // end transfer on last byte
        if ((saved_cmd & CMD_I2C_END) && !expected)
            i2c_stop();

        if (err) {
            //TODO: set status
        }

    } else {
        // not in ack state
        return 0;
    }

    return 1;
}
