/**
 * @file main.c
 *
 * @brief This file contains the main routine with application entry point for
 *        the I2C-MP-USB firmware project.
 *
 * @author Thomas Fischl
 * @copyright (c) 2017-2019 Thomas Fischl
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @section history Change History
 * - v1.0 (2017-09-20)
 *   - First release
 * - v1.1 (2019-07-23)
 *   - Added bootloader start command
 *   - Changed LED signaling
 */

#include <xc.h>
#include "usb.h"
#include "i2c.h"
#include "clock.h"
#include "led.h"
#include "bootloader.h"
#include "i2c_mp_usb.h"

/**
 * @brief Main function and entry point of application
 * @param argc Count of arguments
 * @param argv Argument vector
 * @return Error code
 */
int main(int argc, char** argv) {

    // set clock to 3x PLL, 48 MHz
    OSCCON = 0b11111100;

    // activate active clock tuning with USB clock
    ACTCON = 0b10010000;

    // wait until clock is stable
    while ((OSCSTATbits.HFIOFR == 0) &&
            (OSCSTATbits.PLLRDY == 0));

    ANSELC = 0x00;
    ANSELA = 0x00;

    // set in-/output
    TRISA = 0x00;
    TRISC = 0x03;
    
    // initialize software modules
    clock_init();
    led_init();
    i2c_init();
    usb_init();

    // enable interrupts
    PEIE = 1; // peripheral interrupt enable
    GIE = 1; // enable global interrupts

    // main loop
    while (1) {

        // reset / bootloader
        if (!PORTAbits.RA3) {
            usb_shutdown();
            asm("ljmp 0x000");
        }

        // do module processing
        usb_process();
        bootloader_process();
    };

    return 0;
}

/**
 * @brief Interrupt service routine
 */
void interrupt isr(void) {

    // timer 1 interrupt
    if (TMR1IE && TMR1IF) {
        TMR1IF = 0;

        clock_isr();
        led_isr();
    }

}